/*
 * *      remote868.c  ... simple keeloq remote control @868MHz tunable,
 * *      using RFM119 module or directly CMT2119 chip
 * *
 * *      Copyright (C) 2019-2020 Jiri Pittner <jiri@pittnerovi.com>
 * *
 * *      This program is free software: you can redistribute it and/or modify
 * *      it under the terms of the GNU General Public License as published by
 * *      the Free Software Foundation, either version 3 of the License, or
 * *      (at your option) any later version.
 * *
 * *      This program is distributed in the hope that it will be useful,
 * *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 * *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * *      GNU General Public License for more details.
 * *      You should have received a copy of the GNU General Public License
 * *      along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * *
 * *
 * */

//
//cf. kicad project remote868.pro
//atmega168p/atmega328p xtal 3.6864
//make fuse_atmega168
//NOTE: USART AND ISP must be disconnected otherwise MCU hangs in the sleep mode
//use #define DEBUG to avoid this, but it would drain battery then
//
//        PB1 and PC4 parallel OUT tx_data and twi data
//        PC5 OUT twi clock

//!!!upload also eeprom, LOCK  FLASH
//
#undef debug
#undef DEBUG


/*
PINOUT:
buttons PC0-PC3
TXMOD  PB1
SCL PC5
SDA PC4
PB0 LED
PB2 PWR_HOLD
*/

#define LOCATION bohdalec
//serials 3,4,5
#define REMOTE_SERIAL 5
#define COUNTER_START  1000
#define BAUD 115200
#define TXCOUNT 3
#undef COUNTER32
#define TXONLY



#include "backward.h"
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <setjmp.h>
#include <string.h>
#include <avr/sleep.h>
//obsolete #include <avr/timer.h>
#include <avr/wdt.h>
#include <stdlib.h>
#include "atmega168compat.h"


#define INCFILE_PARAMS_(X) X##_params.h
#define INCFILE_PARAMS(X) INCFILE_PARAMS_(X)
#define INCFILE_KEYS_(X) X##_keys.h
#define INCFILE_KEYS(X) INCFILE_KEYS_(X)
#define XSTR(s) STR(s)
#define STR(s) #s

#define STRLOCATION STR(LOCATION)
#define GEN_REMOTE_ENTRY(x)  XGEN_REMOTE_ENTRY(x)
#define XGEN_REMOTE_ENTRY(x) KEELOQ_GARAZ1_##x

#include XSTR(INCFILE_PARAMS(LOCATION))

#include "keeloq1.h"
#include XSTR(INCFILE_KEYS(LOCATION))

#include "keeloq1.c"


#if defined(at90s2313) || defined(at90s8535)
#else
#define ATmega
#endif

#ifdef ATmega
#define USR UCSRA
#endif


#if (XTAL == 14745600L) 
#define oversampling 4
#else
#define oversampling 2
#endif

#ifndef MCU
#define emulate
#else
#define AVR_assembler
#endif

#define itoa10(N,S) itoa(N,S,10)
#define itoa16(N,S) itoa(N,S,16)



//static uint8_t wear_level[16]  __attribute__ ((section(".eeprom"))) ;
static uint16_t counter_eeprom_output __attribute__ ((section(".eeprom"))) = COUNTER_START;
static uint32_t frequency __attribute__ ((section(".eeprom"))) = REMOTE_FREQUENCY;
static uint16_t deviation __attribute__ ((section(".eeprom"))) = REMOTE_FSK_DEVIATION;
static uint8_t modulation  __attribute__ ((section(".eeprom"))) = REMOTE_MODULATION;


uint32_t atoi10_32(char *p)
{
uint32_t n=0;
while(p && isdigit(*p))
        {
        n*=10;
        n+= *p++ - '0';
        }
return n;
}

char *itoa10_32(uint32_t n,char *p)
{
char *p0=p;
char *q=p;
do {
        *p++ = (n%10) + '0';
        n/=10;
   } while (n);
*p=0;
//reverse order of the string
char z;
--p;
while(p>q) {z=*p; *p=*q; *q=z; --p; ++q;}
return p0;
}


char *itoa16_32(uint32_t n,char *p)
{
char *p0=p;
char *q=p;
do {
	uint8_t x=n&0x0f;
        *p++ = x<10? x+'0': x-10+'a';
        n>>=4;
   } while (n);
*p=0;
//reverse order of the string
char z;
--p;
while(p>q) {z=*p; *p=*q; *q=z; --p; ++q;}
return p0;
}


void putcharx(char c) {loop_until_bit_is_set(USR, UDRE); UDR=c;}

void printP (PGM_P string){
        char c;
        c=pgm_read_byte(string);
                   while (c) {
                   loop_until_bit_is_set(USR, UDRE);
                   UDR = c;
                   c=pgm_read_byte(++string);
                   }
                   return;
                 }



void print (char *string){                                     
                   while (*string) {                                            
                   loop_until_bit_is_set(USR, UDRE);                            
                   UDR = *string++;
                   }                                                            
                   return;                                                      
                 }                                                              


void scan(char *string){
char c;
	do 	{
		do {
			loop_until_bit_is_set(USR, RXC);	
			c =UDR;
			} while bit_is_set(USR, FE);	
		*string++ = c;
		//echo the character
		loop_until_bit_is_set(USR, UDRE);
		UDR = c;
		} while ( c != '\r' );
	loop_until_bit_is_set(USR, UDRE);
        UDR = '\n';
	string[-1]=0;
	}



//UART initialize
#ifdef ATmega
#define UCR UCSRB
#define UART_INIT(baud) { \
UBRRH=0; \
UBRRL= (XTAL/baud+15)/16-1; \
UCSRB=(1<<TXEN)|(1<<RXEN); \
UCSRC=(1<<URSEL)|(1<<UCSZ0)|(1<<UCSZ1)|(1<<USBS); }
#else
#define UART_INIT(baud) { \
UBRR = (XTAL/baud+15)/16-1; \
sbi(UCR, TXEN); \
sbi(UCR, RXEN); \
}
#endif


watchdog(uint8_t onoff)
{
#if(MCU==atmega168 || MCU == atmega168p || MCU == atmega328p)
wdt_enable(WDTO_8S); wdt_reset(); return;
#else
if(onoff) {wdt_enable(WDTO_2S); wdt_reset();}
else {wdt_reset();wdt_disable();}
#endif
}



//transmit routines

volatile uint8_t outcode[MESSAGE_BYTES];
volatile uint8_t outcodelen=0; //0 indicates ready to receive, non-zero indicates ready to read it in main
volatile uint8_t waittime=0;
volatile uint8_t txcount=0; //repeat several times
volatile static uint8_t obitcount=0, obytecount=0; //working counters

typedef enum {init,oupmark,downmark,data,trailer,wait} OSTATE;
static OSTATE ostate=init;
static uint8_t ocount=0;





void txcar(void)
{
switch (ostate) {
	case wait: if(--waittime) return; 
        case init: //the user has just written the request to transmit
			//check maxcodelen
			ocount=13;
			ostate=oupmark;
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p )
			cbi(TIMSK1,TOIE1);
#else
			cbi(TIMSK,TOIE1);
#endif
			ICR1=(uint16_t)(XTAL*8./10000+.5)-1;
                	OCR1A=((uint16_t)(XTAL*4./10000+.5)-1);
			TCNT1=0;
			cbi(TCCR1A,COM1A0);
			cbi(TCCR1A,COM1A1);
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                        sbi(TIMSK1,TOIE1);
                        sbi(TIMSK1,OCIE1A);
#else
			sbi(TIMSK,TOIE1);
			sbi(TIMSK,OCIE1A);
#endif
                break;
        case oupmark:
#if (MCU == atmega8 || MCU == atmega168 || MCU == atmega168p || MCU == atmega328p )
		if(--ocount) {sbi(PORTB,PB1); return;}
		cbi(PORTB,PB1);
#else //ATmega 16 and 32
		if(--ocount) {sbi(PORTD,PD5); return;}
		cbi(PORTD,PD5);
#endif
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
		cbi(TIMSK1,TOIE1);
#else
		cbi(TIMSK,TOIE1);
#endif
                TCNT1=0;
		ICR1=(uint16_t)(XTAL*6.1/10000+.5)-1;
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                sbi(TIMSK1,TOIE1);
#else
                sbi(TIMSK,TOIE1);
#endif
		ostate=downmark;
		ocount=2;
                break;
        case downmark:
		if(--ocount) return;
		obitcount=obytecount=0;
		ostate=data;
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                cbi(TIMSK1,TOIE1);
#else
		cbi(TIMSK,TOIE1);
#endif
		ICR1=(uint16_t)(XTAL*12./10000+.5)-1;
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                sbi(TIMSK1,TOIE1);
#else
                sbi(TIMSK,TOIE1);
#endif
		break;
        case data:
		{
		uint8_t n;
		n=obitcount+(obytecount<<3);
		if(n>outcodelen+1)
                        {
                        //data finished 
#if (MCU == atmega8  || MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                cbi(PORTB,PB1);
#else //ATmega 16 and 32
                cbi(PORTD,PD5);
#endif
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                cbi(TIMSK1,OCIE1A);
#else
		cbi(TIMSK,OCIE1A);
#endif
			ostate=trailer;
			ocount=13;
			OCR1A=0;
                        }
		else
			{
#if (MCU == atmega8  || MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                        if(n) sbi(PORTB,PB1); //avoid spurious first peak
#else //ATmega 16 and 32
                        if(n) sbi(PORTD,PD5); //avoid spurious first peak
#endif
			if(n>outcodelen) 
				{
				OCR1A=((uint16_t)(XTAL*12./10000+.5)-1)*1/3;
				++obitcount;
				}
			else
				{
				OCR1A=((uint16_t)(XTAL*12./10000+.5)-1)*
				((outcode[obytecount]& (1<<obitcount))?1:2)/3;
				if(++obitcount==8) {obitcount=0;++obytecount;}
				}
			}
		}
                break;
        case trailer:
		if(--ocount) return;
		//finished
                if(--txcount == 0) {ostate=init; outcodelen=0;}
                else {ostate=wait; waittime=32;} 
                break;
}
}



INTERRUPT(SIG_OVERFLOW1)
{
if(!outcodelen) return;
txcar();
}


INTERRUPT(SIG_OUTPUT_COMPARE1A)
{
#if (MCU == atmega8  || MCU == atmega168 || MCU == atmega168p || MCU == atmega328p) 
cbi(PORTB,PB1);
#else //ATmega 16 and 32
cbi(PORTD,PD5); 
#endif
}





//delay routines

void delay_xs(uint16_t xs) //xs takes 4 clocks time
{
        asm volatile (
        "\n"
        "L_dl1%=:" "\n\t"
        "sbiw %0, 1" "\n\t"
        "brne L_dl1%=" "\n\t"
        : "=&w" (xs)
        );
        return;
}
__attribute__((noinline));

void delay_ms(uint16_t ms) //ms takes 1/1000 s
{
while(ms--) delay_xs(XTAL/4/1000);
}






void do_transmit(uint8_t onoff)
{
if(onoff)
        {
#if (MCU == atmega8  || MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                sbi(DDRB,PB1); cbi(PORTB,PB1);
		sbi(DDRC,PC5); sbi(PORTC,PC5);
#else //ATmega 16 and 32
		sbi(DDRD,PD7); sbi(PORTD,PD7); 
                cbi(PORTD,PD5); sbi(DDRD,PD5);
#endif
	delay_ms(1);
	TCCR1A= (1<<WGM11);
 	TCCR1B= (1<<WGM12)|(1<<WGM13)|CK;
	ICR1=(uint16_t)(XTAL*12./10000+.5)-1;
	OCR1A=0;
	TCNT1=0;
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
	sbi(TIMSK1,TOIE1);
#else
	sbi(TIMSK,TOIE1);
#endif
	}
else
	{
	TCCR1A=0;
        TCCR1B=0;
#if (MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
        cbi(TIMSK1,TOIE1);
#else
        cbi(TIMSK,TOIE1);
#endif
        cbi(DDRD,PD7); cbi(PORTD,PD7);
#if (MCU == atmega8  || MCU == atmega168 || MCU == atmega168p || MCU == atmega328p)
                cbi(PORTC,PC5); sbi(DDRC,PC5);
                cbi(PORTB,PB1); sbi(DDRB,PB1);
#else //ATmega 16 and 32
                cbi(PORTD,PD5); sbi(DDRD,PD5);
                cbi(PORTD,PD7); sbi(DDRD,PD7);
#endif
	}
}




#define min(x,y) ((x)<(y)?(x):(y))

REMOTE_ENTRY remote_entry= GEN_REMOTE_ENTRY(REMOTE_SERIAL);

int prepare_keeloq(uint8_t buttons, uint8_t *msg)
{
remote_entry.counter = eeprom_read_word(&counter_eeprom_output);
//keeloq counter
printP(PSTR("counter="));
{char z[16]; itoa10(remote_entry.counter,z); print(z);}
printP(PSTR("\n"));
printP(PSTR("buttons="));
{char z[16]; itoa16(buttons,z); print(z);}
printP(PSTR("\n"));

uint8_t len=generate_message(msg,buttons,&remote_entry);
eeprom_write_word(&counter_eeprom_output,remote_entry.counter);
return len;
}


void transmit1(uint8_t *msg, uint8_t len, uint8_t count)
{

if(!count) return;

//wait for end of previous transmit or switch on transmitter
if(txcount || outcodelen)
        {
		do {delay_ms(10);} while(txcount || outcodelen);
        }
else do_transmit(1);


//transmit the code a few times
memcpy(outcode,msg,(len+7)/8);
txcount=count;
outcodelen=len;

//do not wait for end, loop in main checks and starts receiving when finished
}

#define DELAY_ONE delay_xs(XTAL/4*1.21e-3)
#define DELAY_HALF delay_xs(XTAL/4*1.21e-3/2)




//uart input

//UART routines
volatile uint8_t inuartlen=0;
volatile char inuart[64];
INTERRUPT(SIG_UART_RECV)
{
char c=UDR;
if(c=='\r') c='\n';
inuart[inuartlen]=c;
putcharx(c);
inuartlen = (inuartlen+1)&63;
}


//actually it is a kind of simplified TWI, I2C addressing nor acknowledge bits are emplopyed
//therefore we implement it purely in software
void pseudoi2c_on(uint8_t onoff)
{
if(onoff)
	{
	cbi(DDRB,PB1);cbi(PORTB,PB1); //I2C shared with transmittor modulation
	sbi(DDRC,PC4); cbi(PORTC,PC4);
        sbi(DDRC,PC5); sbi(PORTC,PC5);
	}
else
	{
	cbi(DDRC,PC4); cbi(PORTC,PC4);
	cbi(DDRC,PC5); cbi(PORTC,PC5);
	}
}

#define TWI_DELAY 3

void pseudoi2c_reset(void)
{
sbi(PORTC,PC5);
sbi(DDRC,PC4);
cbi(PORTC,PC4);
uint8_t i;
for(i=0; i<32; ++i)
	{
	sbi(PORTC,PC5);
	delay_xs(TWI_DELAY);
	cbi(PORTC,PC5);
        delay_xs(TWI_DELAY);	
	}
}


void pseudoI2C_Write(uint8_t x)
{
uint8_t i;
sbi(PORTC,PC5);
sbi(DDRC,PC4);
cbi(PORTC,PC4);
for(i=0; i<8; ++i)
	{
	sbi(PORTC,PC5);
	if(x&0x80) sbi(PORTC,PC4); else cbi(PORTC,PC4);
	delay_xs(TWI_DELAY);
	cbi(PORTC,PC5);
	delay_xs(TWI_DELAY);
	x<<=1;
	}
sbi(PORTC,PC5);
cbi(PORTC,PC4);
}

uint8_t pseudoI2C_Read(void)
{
uint8_t r=0,i;
cbi(DDRC,PC4);
cbi(PORTC,PC4);
sbi(PORTC,PC5);
for(i=0; i<8; ++i)
	{
	sbi(PORTC,PC5);
	delay_xs(TWI_DELAY);
	r<<=1;
	cbi(PORTC,PC5);
	delay_xs(TWI_DELAY);
	if(bit_is_set(PINC,PC4)) r|=1;
	}
sbi(PORTC,PC5);
sbi(DDRC,PC4);
cbi(PORTC,PC4);
return r;
}


void TWI_WRREG(uint8_t addr, uint8_t data)
{
pseudoI2C_Write(0x80|(addr&0x3f));
pseudoI2C_Write(data);
}

uint8_t TWI_RDREG(uint8_t addr)
{
pseudoI2C_Write(0xc0|(addr&0x3f));
return pseudoI2C_Read();
}


void TWI_RAM1(uint8_t addr, uint16_t data)
{
TWI_WRREG(0x18,addr);
TWI_WRREG(0x19,data&0xff);
TWI_WRREG(0x1A,data>>8);
TWI_WRREG(0x25, 0x01);
}

void TWI_RAM(const uint16_t *x, uint8_t n)
{
uint8_t i;
for(i=0; i<n; ++i) TWI_RAM1(i,*x++);
}


//configuration file generated by proprietary CMOSTEK utility /opt/wine/drive_c/Program Files/CMOSTEK/RFPDK/RFPDK.exe
static const uint16_t CMT2119fsk[21] =
{
0x007F,
0x5000,
0x0000,
0x0000,
0x0000,
0xF000,
0x0000,
0xBB13, //addr.7 - low 16 bits of PLL multiple
0x4208, //addr.8 - high 8 bits of PLL multiple <<8 | 0x08
0x00B0, //addr.9 - deviation with mask 0x07f8?
0x2401,
0x0081,
0x8000,
0x0000,
0xFFFF,
0x0020,
0x5F1E, //related to tx power
0xA2D6,
0x0E13,
0x0019,
0x0000,
};

static const uint16_t CMT2119ook[21] =
{
0x007F,
0x5000,
0x0000,
0x0000,
0x0000,
0xF000,
0x0000,
0xBB13,
0x4200,
0x0000,
0x2401,
0x01B0,
0x8000,
0x0006,
0xFFFF,
0x0020,
0x5F1E,
0x22D6,
0x0E13,
0x0019,
0x2000,
};

//one register remains unclear after a very rudimentary reverse engineering attempt (0x5f1e gets changed at low frequencies - is this critical?
//frequency = 26000000/2^17/prescale*pll_multiple
//prescale can be 1.5, 2, 3 controlled by positions 1 and 6
//gmsk not supported presently
void TWI_FREQ(uint32_t freq, uint32_t dev, uint8_t mod)
{
//determine prescale
uint8_t prescale15;
uint8_t prescale2;
if(freq<=300000000UL)
	{
	prescale15=1;
	prescale2=1;
	}
else if(freq<=400000000UL)
	{
	prescale15=0;
	prescale2=1;
	}
else if(freq<=600000000UL)
        {
        prescale15=1;
        prescale2=0;
        }
else
	{
	prescale15=0;
        prescale2=0;
	}

if(prescale15) TWI_RAM1(6,0x0001);
if(prescale2) TWI_RAM1(1,0x5400);

int8_t psc2=0;
if(prescale15) psc2-= 1;
if(prescale2) psc2+=1;

//frequency
uint64_t tmp1=freq;
tmp1 <<=17+psc2;
if(prescale15) tmp1*=3;
uint32_t pll = tmp1/26000000UL;
TWI_RAM1(7,pll&0xffff);
TWI_RAM1(8,((pll>>16)<<8)|0x08);
/*
{
printP(PSTR("PLL_freq="));
char z[16];
itoa16_32(pll,z);
print(z);
printP(PSTR("\n"));
}
*/

if(mod==2)
	{
	//deviation
	tmp1=dev;
	tmp1 <<=14+psc2;
	if(prescale15) tmp1*=3;
	pll = tmp1/26000000UL;
	pll<<= 3;
	pll &=  0x07f8;
	}
else pll=0; //no deviation in ook
TWI_RAM1(9,pll);

/*
{
printP(PSTR("PLL_dev="));
char z[16];
itoa16_32(pll,z);
print(z);
printP(PSTR("\n"));
}
*/

}


//this configures only the RAM registers of CMT2119, not its EEPROM
//programming EEPROM is not documented, can it actually be done from the MCU?
//original Cmostek USB programmer is rather expensive...
//
void configure(uint32_t freq, uint16_t dev, uint8_t mod)
{
cbi(SREG,7);
uint16_t *p;
uint8_t i;
pseudoi2c_on(1);

pseudoi2c_reset(); //step 1
TWI_WRREG(0x3d, 0x01); //step 2
delay_ms(2);

//some proprietary command preamble from the datasheet
TWI_WRREG(0x02, 0x78); //Open LDO & Osc step 3

TWI_WRREG(0x2F, 0x80); //vActiveRegsister step 4
TWI_WRREG(0x35, 0xCA);
TWI_WRREG(0x36, 0xEB);
TWI_WRREG(0x37, 0x37);
TWI_WRREG(0x38, 0x82);

TWI_WRREG(0x12, 0x10); //vEnableRegMode step 5
TWI_WRREG(0x12, 0x00);
TWI_WRREG(0x24, 0x07);
TWI_WRREG(0x1D, 0x20);

//program the RAM config by RFPDK generated setup
TWI_RAM((mod==2?CMT2119fsk:CMT2119ook),21); //step 6

//reprogram frequency
TWI_FREQ(freq,dev,mod);

TWI_WRREG(0x0D, 0x02); //step 7

cbi(PORTC,PC4);
delay_ms(2);

pseudoi2c_on(0);
sbi(SREG,7);
}



void led(uint8_t onoff)
{
if(onoff) sbi(PORTB,PB0);
else cbi(PORTB,PB0);
}


void bin2letter(char *c, uint8_t b)
{
*c = b<10? '0'+b : 'A'+b-10;
}

void bin2octet(char *octet, uint8_t bin)
{
bin2letter(octet,bin>>4);
bin2letter(octet+1,bin&0x0f);
}

void printmsg(uint8_t *msg, uint8_t len)
{
printP(PSTR("T: "));
char c[4];
bin2octet(c,len);
c[2]=0;
print(c); printP(PSTR(" "));
uint8_t i;
for(i=0; i<(len+7)>>3; ++i)
                {
                bin2octet(c,msg[i]);
                c[2]=0;
                print(c);
                }
}

//ADC

int32_t adcread(uint8_t mux, uint16_t nrep)
{
ADCSRA=1<<ADEN;
ADMUX= mux;
delay_ms(1); //let amplifiers and adc input settle
cbi(SREG,7);
int32_t r=0;
uint16_t i;
for(i=0; i<=nrep; ++i)
        {
        int16_t v;
        ADCSRA= (1<<ADEN)|(1<<ADIF)|(1<<ADSC)|6;
        loop_until_bit_is_set(ADCSRA,ADIF);
        v=ADCL;
        v|= (ADCH<<8);
        //if(i<=16) {char c[20]; sprintf(c,"ADC 0x%02x %d\n",mux,v);print(c);}
        if(i>0) //skip the first measurement after channel change to let it stabilize
                {
                if((mux&0x1f)>=8 && v>=512) v-=1024; //differential
                r += v;
                }
        }
ADCSRA=0;
sbi(SREG,7);
return r;
}

#define VOLTAGE_DIVISOR ((22000.f+10000.f)/10000.f)
#define V_REFERENCE 1.1f
uint16_t voltage(void) //integer, in mV
{
uint32_t v;
v=adcread(6|(1<<REFS0)|(1<<REFS1),16);
float u= v*V_REFERENCE/1024.*VOLTAGE_DIVISOR/16;
return (uint16_t) 1000*u;
}



int main(void)
{
//grab power
sbi(DDRB,PB2);
sbi(PORTB,PB2);

//init TX pins
pseudoi2c_on(0);
do_transmit(0);

//init buttons
cbi(DDRC,PC0);
cbi(DDRC,PC1);
cbi(DDRC,PC2);
cbi(DDRC,PC3);


//init led
sbi(DDRB,PB0); cbi(PORTB,PB0);

UART_INIT(BAUD);
{char dummy=UDR;}
sbi(UCSRB,RXCIE);
printP(PSTR("Reset!\n"));

while(1)
{

sbi(SREG,7); //global interrupt activate

//read buttons to see if any is pressed
uint8_t buttonsraw= PINC&0x0f;
if(buttonsraw==0)  goto powerdown;

//configure CMT2119 to required setup
configure(eeprom_read_dword(&frequency),eeprom_read_word(&deviation),eeprom_read_byte(&modulation));

//let the user press two buttons simultaneously - re-read buttons again after a moment
delay_ms(20);
buttonsraw= PINC&0x0f;


//map the butttons to backward compatibility with older remote
uint8_t buttons=0;
if(buttonsraw&1) buttons |= BUTTON_I;
if(buttonsraw&2) buttons |= BUTTON_II;
if(buttonsraw&4) buttons |= BUTTON_III;
if(buttonsraw&8) buttons |= BUTTON_IV;

//prepare meessage to broadcast
uint8_t msg[MESSAGE_BYTES];
uint8_t lastbuttons=buttons;
uint8_t len=prepare_keeloq(buttons,msg);

//transmit while buttons pressed
while(buttons==lastbuttons && buttons!=0)
	{
        printmsg(msg,len);
	transmit1(msg, len, TXCOUNT);
	while(outcodelen);//wait for end of transmition
	printP(PSTR(" ...done\n"));
	led(1);
	delay_ms(100);
	led(0);
	delay_ms(100);
	lastbuttons=buttons;
	buttons= PINC&0x0f;
	}
do_transmit(0);

if(buttons==0) goto powerdown;

//setup mode via uart
led(1);
delay_ms(800);
led(0);

printP(PSTR("entering setup mode\n"));

uint16_t timeout=1000; //>=10s 
while(1)
	{
	if(--timeout==0) break;
	if(inuart[inuartlen-1]== '\n')
   		{
		timeout=1000;
   		inuart[inuartlen]=0;
   		inuart[inuartlen-1] = '\n';
		//execute the command
		switch(inuart[0]) 
			{
			case 'f': 
				{
				uint32_t freq=atoi10_32(inuart+1);
				eeprom_write_dword(&frequency,freq);
				}
				break;
			case 'F': 
                                {
                                uint16_t dev=atoi(inuart+1);
                                eeprom_write_word(&deviation,dev);
                                }
                                break;
			case 'm':
				{
				uint8_t mod=inuart[1]-'0';
				eeprom_write_byte(&modulation,mod);
				}
				break;
			case 'c': //show counter
				printP(PSTR("counter="));
				{char z[16]; itoa10(eeprom_read_word(&counter_eeprom_output),z); print(z);}
				printP(PSTR("\n"));
				break;
			case 'C': //set counter
				{
				uint16_t c=atoi(inuart+1);
				eeprom_write_word(&counter_eeprom_output,c);
				}				
				break;
			case 'T': //test transmission
				{
				pseudoi2c_on(0);
				do_transmit(1); 
				uint16_t i;
				for(i=0; i<10000; ++i) {sbi(PORTB,PB1);delay_ms(1);  cbi(PORTB,PB1); delay_ms(1);}
				do_transmit(0);
				printP(PSTR("done\n"));
				}
				break;
			case 'W': //cw transmission
				{
				pseudoi2c_on(0);
                                do_transmit(1); 
				sbi(PORTB,PB1);
				delay_ms(10000);
				cbi(PORTB,PB1);
				do_transmit(0);
				printP(PSTR("done\n"));
				}
				break;
			case 'q': 
				goto powerdown;
				break;
			default: 
				printP(PSTR("Unknown command: "));
                		print(inuart);
                		break;
			}
		inuartlen=0;
		}

	delay_ms(10);
	}

powerdown: ;

//indicate weak battery
{
uint16_t v=voltage();// in mV
printP(PSTR("voltage="));
{char z[16]; itoa10(v,z); print(z);}
printP(PSTR("\n"));
if(v<2000)
	{
	uint8_t i,j;
	for(i=0; i<=20; ++i) 
	    for(j=0; j<10; ++j)
		{
		sbi(PORTB,PB0);
		delay_ms(20-i);
		cbi(PORTB,PB0);
		delay_ms(i);
		}
	}
}

#ifndef DEBUG
printP(PSTR("power down\n"));
delay_ms(5);
cbi(DDRC,PC4);cbi(PORTC,PC4);
cbi(DDRC,PC5);cbi(PORTC,PC5);
cbi(DDRB,PB1);cbi(PORTB,PB1);

//LED off
cbi(PORTB,PB0);
cbi(DDRB,PB0);

//power down UART
cbi(UCR, TXEN);
cbi(UCR, RXEN);

//power hold off
cbi(PORTB,PB2);
cbi(DDRB,PB2);
set_sleep_mode(SLEEP_MODE_PWR_DOWN);
sleep_mode();
#endif
//DEBUG

}//while executed only in DEBUG mode
}
